#!/usr/bin/env python
from __future__ import (print_function as _pf,
                        unicode_literals as _ul,
                        absolute_import as _ai)


def outlier_bounds_std(arr, multiplier=3):
    """
    Get the upper and lower limits for outlier detection
    using mean and standard deviation:
        mean +- std * [3]
    the multiplier [3] can be adjusted by the user
    returns the lower_limit, upper_limit
    """
    from numpy import array, nanmean, nanstd
    arr = array(arr)

    mean = nanmean(arr)
    std = nanstd(arr)

    ll = mean - std * multiplier
    ul = mean + std * multiplier

    return ll, ul


def outlier_bounds_iqr(arr, multiplier=1.5):
    """
    Get the upper and lower limits for outlier detection
    using interquartile range:
        LL = Q1 - IQR * [1.5]
        UL = Q3 + IQR * [1.5]

    the multiplier [1.5] can be adjusted by the user
    returns the lower_limit, upper_limit
    """
    from numpy import array, nanpercentile
    arr = array(arr)

    q1, q3 = nanpercentile(arr, [25, 75])
    iqr = q3 - q1

    ll = q1 - iqr * multiplier
    ul = q3 + iqr * multiplier

    return ll, ul


def horizontal_diff_outliers(dives, depth, arr, depth_threshold=450, mask_frac=0.2):
    """
    Find outliers below a the threshold depth.
    Uses the horizontal gradient as a threshold to find outliers.
    Useful when a dive at depth is not the same as neighbouring values
    """
    from numpy import array, nanmean, nanstd, abs, inf
    from . tools import grid_data

    dives = array(dives)

    # grid data so that the horizontal rolling median can be calculated
    # we use a window of 3 to find only "horizonal spikes"
    gridded = grid_data(dives, depth, array(arr))
    median = gridded.rolling(3, axis=1, center=True).median()
    # get zscore of the difference between the median and the raw data
    diff = gridded - median
    zdiff = abs(diff - nanmean(diff)) / nanstd(diff)

    # this finds the 99.7th percentile outliers
    # note that this is based on the global horizonal diff
    # but is only applied below the depth threshold
    # this means that the surface data sets a higher limit
    deep_outlier = zdiff.loc[depth_threshold:] >= 3

    # get the ratio of bad values per dive and mask if it
    # exceeds a user defined fraction
    deep_outlier_count = deep_outlier.sum()
    deep_obs_num = gridded.shape[0] - depth_threshold  # assumes bin of 1m
    deep_outlier_ratio = deep_outlier_count / deep_obs_num
    # finds the index where dives exceed the mask_frac threshold
    i = deep_outlier_ratio > mask_frac
    deep_outlier_dives = i[i].index.values

    mask = arr < -inf  # create a dummy mask
    for d in deep_outlier_dives:
        i = dives == d
        mask[i] = True

    return mask


def mask_bad_dive_fraction(mask, dives, mask_frac=0.2):
    """
    Pass a mask array to this function to find bad dives - where more
    than a fraction of the dive is masked

    INPUT:
        mask - boolean 1D array with masked values
        dives - discrete dive numbers (down round, up n.5)
        mask_frac = fraction of the dive that is masked for the whole dive to be bad
    """
    from pandas import Series
    from numpy import array
    # catch dives where the marjority of the data is masked
    # and return a fully masked dive
    grp = Series(mask).groupby(dives.values)
    masked_frac_per_dive = grp.sum() / grp.count() > mask_frac
    majority_masked = masked_frac_per_dive[masked_frac_per_dive].index.values

    # create a mask that masks ungridded data
    mask_dives = mask.copy()
    for d in majority_masked:
        i = array(dives) == d
        mask_dives[i] = True
    return mask_dives


def blob_outliers(x, y, conv_matrix=None, min_count=5, verbose=True):
    """
    Use a blob detection method to find outliers based on two variables.
    NOTE that this outlier detection method needs tuning to work well.
    This uses convolution to create the blob - you can specify the convolution matrix.

    INPUT:
        x, y - same length vectors can be pandas.Series or numpy.arrays
        conv_matrix   Can be int, [int, int] or 2d array
                      int = size of the isotropic round convolution window
                      [int, int] = anisotropic (oval) convoltion window
                      2d array is a weighted convolution window
                      rectangle = np.ones([int, int])
                      oval = bgu.tools.gaussian_kernel([int, int]) > 1e-5
                      more advanced anisotropic windows can also be created
        min_count     masks the 2d histogram counts smaller than this limit
                      when performing the convolution - default is 1
        verbose       returns figures - default is True
    OUPUT:
        mask - same shape as x or y
        fig - only returned if verbose is True

    SUGGESTIONS:
        x = salinity, y = temperature
    """
    from scipy.signal import convolve2d
    from pandas import Series, cut
    from numpy import linspace, c_, where, inf, array, isnan
    from . tools import gaussian_kernel

    # turning input into pandas.Series
    x = Series(x, name='X' if not isinstance(x, Series) else x.name)
    y = Series(y, name='Y' if not isinstance(y, Series) else y.name)

    ###############
    ##  BINNING  ##
    ###############
    # create bins for the data - equal bins
    xbins = linspace(x.min(), x.max(), 250)
    ybins = linspace(y.min(), y.max(), 250)
    # binning the data with pandas. This is quick to find outliers at the end
    xcut = cut(x, xbins, labels=c_[xbins[:-1], xbins[1:]].mean(1), right=False)
    ycut = cut(y, ybins, labels=c_[ybins[:-1], ybins[1:]].mean(1), right=False)

    # binning the data and returning as a 2D array (pandas.DataFrame)
    count = x.groupby([xcut, ycut]).count()
    count.name = 'count'  # to avoid an error when unstacking
    count = count.unstack()
    count = count.sort_index().sort_index(axis=1)

    ###################
    ##  CONVOLUTION  ##
    ###################
    # make convolution matrix if not given
    if conv_matrix is None:
        conv_matrix = (gaussian_kernel(21) > 1e-5).astype(int)
    elif isinstance(conv_matrix, (list, int, float)):
        conv_matrix = (gaussian_kernel(conv_matrix) > 1e-5).astype(int)
    else:
        ndim = array(conv_matrix).ndim
        if ndim != 2:
            raise UserWarning('conv_matrix must have 2 dimensions')
    # An array with which the convolution is done
    # use a threshold to mask out bins with low counts
    # thus only dense regions of data are considered
    count0 = count.fillna(0).values
    count0[count0 < min_count] = 0
    # 2d convolution with the input matrix
    convolved_count = convolve2d(count0, conv_matrix, mode='same')
    outliers = (convolved_count == 0) & ~isnan(count)

    cols = count.index
    rows = count.columns

    ########################################
    ##  FINDING OUTLIERS AND CREATE MASK  ##
    ########################################
    # find indicies of of the where there is no convolution,
    # but there are data. Then get the x and y values of these
    # points. Turn these into pairs for pandas multi-indexing.
    i, j = where(outliers)
    xi = cols[i].values
    yj = rows[j].values
    ij = list(zip(xi, yj))
    # Create a pandas dataframe with the pd.cut data as indicies
    # with a column for a numerical index.
    if len(ij) > 0:
        idx = x.to_frame().reset_index().drop(x.name, axis=1)
        idx = idx.set_axis([xcut, ycut], inplace=False)
        idx = idx.loc[ij]['index'].values
    else:
        idx = None
    # create a placeholder mask and fill outliers with True
    mask = (x > inf).values
    mask[idx] = True

    ###############
    ##  FIGURES  ##
    ###############
    if verbose:
        from numpy import ma, power, diff, sum, r_
        from matplotlib import colors, cm, pyplot as plt
        # x and y plotting coordinates
        xp = cols.values.astype(float)
        yp = rows.values.astype(float)
        # plotting variables a, b, c
        a = ma.masked_invalid(count.T, 0)
        b = convolved_count.T
        c = ma.masked_where(a.mask, ~outliers.T)

        # create the figure
        fig, ax = plt.subplots(1, 2, figsize=[10, 5], dpi=90, sharey=True)
        # properties for the pcolormesh and contours
        pn = colors.PowerNorm(0.3)
        mesh_props = dict(cmap=cm.Spectral_r, norm=pn)
        # create the pcolormesh plots
        im = (ax[0].pcolormesh(xp, yp, a, vmax=a.max()/2, **mesh_props),
              ax[1].pcolormesh(xp, yp, c, vmin=0, vmax=1))
        ct = ax[1].contour(xp, yp, b, levels=[0.5], linestyles='-', colors='r', linewidths=2)

        # change figure parameters
        ax[0].set_title('Histogram of data (min_count = {})'.format(min_count))
        ax[1].set_title('{} Outliers found using \n{} convolution with decision boundary'.format(mask.sum(), str(conv_matrix.shape)))
        ax[0].set_xticks([])
        ax[0].set_ylabel(y.name)
        ax[1].set_xlabel(x.name)

        # tight layout before creating the axes for pcolomesh plots
        fig.tight_layout()

        # make colorbar axes based on axes [0, 1]
        p = ax[0].get_position()
        cax = fig.add_axes([p.x0, p.y0-0.05, p.width, 0.04])
        cb = plt.colorbar(im[0], cax=cax, orientation='horizontal')
        cb.set_label('Count')
        cb.set_ticks([1, 2, 3, 5, 10, 30, 80, 200])
        # plot the min_count on the colorbar
        cx = pn(r_[cb.get_clim(), min_count])[-1]
        cb.ax.plot(cx, 0, marker='^', color='k', markersize=8)
        cb.ax.plot(cx, 1, marker='v', color='k', markersize=8)
        return mask, fig

    return mask


def despike(arr, window_size, spike_method='median'):
    """
    This script is copied from Nathan Briggs' MATLAB script using in Briggs et al (2011).
    It returns the baseline of the data and the residuals of
    [measurements - baseline]

    INPUT ARGUMENTS
        arr          a pandas.Series or numpy.ndarray
        window_size  the length of the rolling window size
        method       a string with `minmax` or `median`

    OUTPUT
        baseline     the baseline from which noise limits are determined
        spikes       spikes are the residual of [measurements - baseline ]

    NOTES
        method='minmax' first applies a rolling minimum to the dataset
            thereafter a rolling maximum is applied. This forms the baseline.
            Spikes are the difference from the baseline
        method='median' is a rolling median applied to the dataset
            spikes are the difference between median and baseline, thus more
            likely to be negative
    """
    from numpy import array, ndarray, isnan, nanmedian, nanmin, nanmax, nan

    # convert to array
    arr = array(arr)
    # create empty array for baseline
    baseline = ndarray(arr.shape) * nan
    # mask with exisiting nans masked out
    mask = ~isnan(arr)

    # if min-max method then get the rolling minimum and
    # then the rolling maximum
    if spike_method.startswith('min'):
        base_min = rolling_window(arr[mask], nanmin, window_size)
        base = rolling_window(base_min, nanmax, window_size)
    else:
        base = rolling_window(arr[mask], nanmedian, window_size)

    baseline[mask] = base
    spikes = arr - baseline

    return baseline, spikes


def despiking_report(dives, depth, raw, baseline, spikes, name=None, pcolor_kwargs={}):
    """
    Creates a figure showing the difference between the baseline estimate and spikes

    INPUT:
        dives, depth, raw - same length vectors that are also passed to the despike function
        baseline, spikes  - output from the despike function
        name              = string name for the figure header.
                            Will try to get a pandas.Series.name from raw
        pcolor_kwargs     = a dictionary of keyword arguements passed to pcolormesh

    OUTPUT:
        figure object
    """
    from matplotlib.pyplot import subplots, cm
    from numpy import array, unique, nanpercentile, ma, isnan
    from . import plot

    if name is None:
        name = 'Variable' if not hasattr(raw, 'name') else raw.name

    x = array(dives)
    y = array(depth)
    z = [array(raw),
         ma.masked_array(baseline),
         array(spikes)]

    fig, ax = subplots(3, 1, figsize=[10, 11])
    title = '{}\nDespiking Report'.format(name)

    bmin, bmax = nanpercentile(z[1].data, [2, 98])
    smin, smax = nanpercentile(z[2].data, [2, 98])
    props = dict(cmap=cm.Spectral_r)
    props.update(pcolor_kwargs)

    im = []
    im += plot.pcolormesh(x, y, z[0], ax=ax[0], vmin=bmin, vmax=bmax, **props),
    im += plot.pcolormesh(x, y, z[1], ax=ax[1], vmin=bmin, vmax=bmax, **props),
    im += plot.pcolormesh(x, y, z[2], ax=ax[2], vmin=smin, vmax=smax, **props),

    for i in range(0, 3):
        ymax = y[~isnan(baseline)].max()
        ax[i].set_ylim(ymax, 0)
        ax[i].set_xlim(x.min(), x.max())
        ax[i].set_ylabel('Depth (m)')
        if i != 2:
            ax[i].set_xticklabels([])
        else:
            ax[i].set_xlabel('Dive number')
        ax[i].cb.set_label('Units')

    ax[0].set_title('Original')
    ax[1].set_title('Despiked')
    ax[2].set_title('Spikes')

    fig.tight_layout()
    fig.text(0.47, 1.02, title, va='center', ha='center', size=14)

    p0 = ax[0].get_position()
    p1 = ax[1].get_position()
    ax[0].set_position([p0.x0, p0.y0, p1.width, p0.height])

    return fig


def rolling_window(arr, func, window):
    """
    A rolling window function that is nan-resiliant and returns
    the full length of the input (pandas and convolve do not do this)
    It applies a given aggregating function to the window.
    """
    from numpy import ndarray, array, nan, r_
    n = window
    # create an empty 2D array with shape (window, len(arr))
    mat = ndarray([n, len(arr) - n]) * nan
    # create a vector for each window
    for i in range(n):
        mat[i, :] = arr[i: i - n]
    # get the mean or meidan or any other function of the matrix
    out = func(mat, axis=0)

    # the array will be shorter than the original
    # pad the output with the rolling average of the values left out
    i0 = n // 2
    i1 = n - i0
    seg0 = array([func(arr[: i + 1]) for i in range(i0)])
    seg1 = array([func(arr[-i - 1:]) for i in range(i1)])
    out = r_[seg0, out, seg1]

    return out


def savitzky_golay(arr, window_size, order, deriv=0, rate=1, interpolate=True):
    """
    Smooth (and optionally differentiate) data with a Savitzky-Golay filter.
    The Savitzky-Golay filter removes high frequency noise from data.
    It has the advantage of preserving the original shape and
    features of the signal better than other types of filtering
    approaches, such as moving averages techniques.
    NOTE: By default, nans in the array are interpolated with a limit set to
          the window size of the dataset before smoothing. The nans are
          inserted back into the dataset after the convolution. This limits
          the loss of data over blocks where there are nans. This can be
          switched off with the `interpolate` keyword arguement.

    Parameters
    ----------
    y : array_like, shape (N,)
        the values of the time history of the signal.
    window_size : int
        the length of the window. Must be an odd integer number.
    order : int
        the order of the polynomial used in the filtering.
        Must be less then `window_size` - 1.
    deriv: int
        the order of the derivative to compute (default = 0 means only smoothing)
    Returns
    -------
    ys : ndarray, shape (N)
        the smoothed signal (or it's n-th derivative).
    Notes
    -----
    The Savitzky-Golay is a type of low-pass filter, particularly
    suited for smoothing noisy data. The main idea behind this
    approach is to make for each point a least-square fit with a
    polynomial of high order over a odd-sized window centered at
    the point.
    Examples
    --------
    t = linspace(-4, 4, 500)
    y = exp( -t**2 ) + random.normal(0, 0.05, t.shape)
    ysg = savitzky_golay(y, window_size=31, order=4)
    import matplotlib.pyplot as plt
    plt.plot(t, y, label='Noisy signal')
    plt.plot(t, exp(-t**2), 'k', lw=1.5, label='Original signal')
    plt.plot(t, ysg, 'r', label='Filtered signal')
    plt.legend()
    plt.show()
    References
    ----------
    .. [1] A. Savitzky, M. J. E. Golay, Smoothing and Differentiation of
       Data by Simplified Least Squares Procedures. Analytical
       Chemistry, 1964, 36 (8), pp 1627-1639.
    .. [2] Numerical Recipes 3rd Edition: The Art of Scientific Computing
       W.H. Press, S.A. Teukolsky, W.T. Vetterling, B.P. Flannery
       Cambridge University Press ISBN-13: 9780521880688
    """
    from math import factorial
    from numpy import array, abs, mat, linalg, concatenate, convolve, isnan, nan
    from pandas import Series

    # sorting out window stuff
    try:
        window_size = abs(int(window_size))
        order = abs(int(order))
    except ValueError:
        raise ValueError("window_size and order have to be of type int")
    if window_size % 2 != 1 or window_size < 1:
        raise TypeError("window_size size must be a positive odd number")
    if window_size < order + 2:
        raise TypeError("window_size is too small for the polynomial order")
    order_range = range(order + 1)
    half_window = (window_size - 1) // 2

    # allow to interpolate for the window size
    if interpolate:
        ser = Series(arr).interpolate(limit=window_size)
        y = array(ser)
    else:
        y = array(arr)

    # precompute coefficients
    b = mat([[k**i for i in order_range] for k in range(-half_window, half_window + 1)])
    m = linalg.pinv(b).A[deriv] * rate**deriv * factorial(deriv)
    # pad the signal at the extremes with
    # values taken from the signal itself
    firstvals = y[0] - abs(y[1:half_window + 1][::-1] - y[0])
    lastvals = y[-1] + abs(y[-half_window - 1:-1][::-1] - y[-1])
    y = concatenate((firstvals, y, lastvals))

    smoothed = convolve(m[::-1], y, mode='valid')

    oldnans = isnan(arr)
    smoothed[oldnans] = nan

    return smoothed
